use crate::db::{self, *};

use sea_orm::{sea_query::IntoCondition, *};
use sea_query::{Alias, Expr, Query, SelectStatement};
use serde::Deserialize;

/// How many fields may be inserted at once into the database.
const PACKAGE_INSERT_LIMIT: usize = 1000;

#[derive(Deserialize)]
pub struct Filter {
    repo: Option<i32>,
    arch: Option<String>,
    name: Option<String>,
}

impl IntoCondition for Filter {
    fn into_condition(self) -> Condition {
        Condition::all()
            .add_option(self.repo.map(|repo| package::Column::RepoId.eq(repo)))
            .add_option(self.arch.map(|arch| package::Column::Arch.eq(arch)))
            .add_option(self.name.map(|name| package::Column::Name.contains(name)))
    }
}

pub async fn page(
    conn: &DbConn,
    per_page: u64,
    page: u64,
    filter: Filter,
) -> crate::Result<Vec<package::Model>> {
    let p2 = Alias::new("p2");
    let query = Query::select()
        .columns(db::package::Column::iter().map(|c| (db::package::Entity, c)))
        .from(db::package::Entity)
        .join_subquery(
            JoinType::InnerJoin,
            max_pkg_ids_query(true),
            p2.clone(),
            Expr::col((db::package::Entity, db::package::Column::Id))
                .eq(Expr::col((p2.clone(), Alias::new("max_id")))),
        )
        .cond_where(filter)
        .order_by((db::package::Entity, db::package::Column::Id), Order::Asc)
        .to_owned();
    let builder = conn.get_database_backend();
    let sql = builder.build(&query);

    Ok(db::Package::find()
        .from_raw_sql(sql)
        .paginate(conn, per_page)
        .fetch_page(page)
        .await?)
}

pub async fn by_id(conn: &DbConn, id: i32) -> Result<Option<package::Model>> {
    package::Entity::find_by_id(id).one(conn).await
}

pub async fn by_fields(
    conn: &DbConn,
    repo_id: i32,
    name: &str,
    version: &str,
    arch: &str,
    compression: &str,
) -> Result<Option<package::Model>> {
    let cond = Condition::all()
        .add(package::Column::RepoId.eq(repo_id))
        .add(package::Column::Name.eq(name))
        .add(package::Column::Arch.eq(arch))
        .add(package::Column::Version.eq(version))
        .add(package::Column::Compression.eq(compression));

    Package::find().filter(cond).one(conn).await
}

pub async fn delete_with_arch(conn: &DbConn, repo_id: i32, arch: &str) -> Result<DeleteResult> {
    Package::delete_many()
        .filter(package::Column::RepoId.eq(repo_id))
        .filter(package::Column::Arch.eq(arch))
        .exec(conn)
        .await
}

pub async fn insert(
    conn: &DbConn,
    repo_id: i32,
    pkg: crate::repo::package::Package,
) -> Result<package::Model> {
    let info = pkg.info;

    // Doing this manually is not the recommended way, but the generic error type of the
    // transaction function didn't play well with my current error handling
    let txn = conn.begin().await?;

    let model = package::ActiveModel {
        id: NotSet,
        repo_id: Set(repo_id),
        base: Set(info.base),
        name: Set(info.name),
        version: Set(info.version),
        arch: Set(info.arch),
        size: Set(info.size),
        c_size: Set(info.csize),
        description: Set(info.description),
        url: Set(info.url),
        build_date: Set(info.build_date),
        packager: Set(info.packager),
        pgp_sig: Set(info.pgpsig),
        pgp_sig_size: Set(info.pgpsigsize),
        sha256_sum: Set(info.sha256sum),
        compression: Set(pkg.compression.extension().unwrap().to_string()),
        state: Set(PackageState::PendingCommit),
    };

    let pkg_entry = model.insert(&txn).await?;

    // Insert all the related tables
    PackageLicense::insert_many(info.licenses.iter().map(|s| package_license::ActiveModel {
        package_id: Set(pkg_entry.id),
        name: Set(s.to_string()),
    }))
    .on_empty_do_nothing()
    .exec(&txn)
    .await?;

    PackageGroup::insert_many(info.groups.iter().map(|s| package_group::ActiveModel {
        package_id: Set(pkg_entry.id),
        name: Set(s.to_string()),
    }))
    .on_empty_do_nothing()
    .exec(&txn)
    .await?;

    let related = info
        .conflicts
        .iter()
        .map(|s| (PackageRelatedEnum::Conflicts, s))
        .chain(
            info.replaces
                .iter()
                .map(|s| (PackageRelatedEnum::Replaces, s)),
        )
        .chain(
            info.provides
                .iter()
                .map(|s| (PackageRelatedEnum::Provides, s)),
        )
        .chain(info.depends.iter().map(|s| (PackageRelatedEnum::Depend, s)))
        .chain(
            info.makedepends
                .iter()
                .map(|s| (PackageRelatedEnum::Makedepend, s)),
        )
        .chain(
            info.checkdepends
                .iter()
                .map(|s| (PackageRelatedEnum::Checkdepend, s)),
        )
        .chain(
            info.optdepends
                .iter()
                .map(|s| (PackageRelatedEnum::Optdepend, s)),
        );
    let related = crate::util::Chunked::new(related, PACKAGE_INSERT_LIMIT);

    for chunk in related {
        PackageRelated::insert_many(
            chunk
                .into_iter()
                .map(|(t, s)| package_related::ActiveModel {
                    package_id: Set(pkg_entry.id),
                    r#type: Set(t),
                    name: Set(s.to_string()),
                }),
        )
        .on_empty_do_nothing()
        .exec(&txn)
        .await?;
    }

    let files = crate::util::Chunked::new(pkg.files, PACKAGE_INSERT_LIMIT);

    for chunk in files {
        PackageFile::insert_many(chunk.into_iter().map(|s| package_file::ActiveModel {
            package_id: Set(pkg_entry.id),
            path: Set(s.display().to_string()),
        }))
        .on_empty_do_nothing()
        .exec(&txn)
        .await?;
    }

    txn.commit().await?;

    Ok(pkg_entry)
}

pub async fn full(conn: &DbConn, id: i32) -> Result<Option<FullPackage>> {
    if let Some(entry) = by_id(conn, id).await? {
        let licenses: Vec<String> = entry
            .find_related(PackageLicense)
            .select_only()
            .column(package_license::Column::Name)
            .into_tuple()
            .all(conn)
            .await?;
        let groups: Vec<String> = entry
            .find_related(PackageGroup)
            .select_only()
            .column(package_group::Column::Name)
            .into_tuple()
            .all(conn)
            .await?;
        let related: Vec<(db::PackageRelatedEnum, String)> = entry
            .find_related(PackageRelated)
            .select_only()
            .columns([package_related::Column::Type, package_related::Column::Name])
            .into_tuple()
            .all(conn)
            .await?;
        let files: Vec<String> = entry
            .find_related(PackageFile)
            .select_only()
            .column(package_file::Column::Path)
            .into_tuple()
            .all(conn)
            .await?;

        Ok(Some(FullPackage {
            entry,
            licenses,
            groups,
            related,
            files,
        }))
    } else {
        Ok(None)
    }
}

#[derive(FromQueryResult)]
pub struct PkgToRemove {
    pub repo_id: i32,
    pub id: i32,
}

fn max_pkg_ids_query(committed: bool) -> SelectStatement {
    let mut query = Query::select()
        .from(db::package::Entity)
        .columns([
            db::package::Column::RepoId,
            db::package::Column::Arch,
            db::package::Column::Name,
        ])
        .expr_as(db::package::Column::Id.max(), Alias::new("max_id"))
        .group_by_columns([
            db::package::Column::RepoId,
            db::package::Column::Arch,
            db::package::Column::Name,
        ])
        .to_owned();

    if committed {
        query.cond_where(db::package::Column::State.eq(db::PackageState::Committed));
    }

    query
}

/// Query that returns all packages that should be included in a sync for the given repository and
/// arch.
pub fn pkgs_to_sync(
    conn: &DbConn,
    repo: i32,
    arch: &str,
) -> SelectorRaw<SelectModel<package::Model>> {
    let (p1, p2) = (Alias::new("p1"), Alias::new("p2"));
    let query = Query::select()
        .columns(db::package::Column::iter().map(|c| (p1.clone(), c)))
        .from_as(db::package::Entity, p1.clone())
        .join_subquery(
            JoinType::InnerJoin,
            max_pkg_ids_query(false),
            p2.clone(),
            Expr::col((p1.clone(), db::package::Column::Id))
                .eq(Expr::col((p2.clone(), Alias::new("max_id")))),
        )
        .cond_where(
            Condition::all()
                .add(Expr::col((p1.clone(), db::package::Column::RepoId)).eq(repo))
                .add(
                    Expr::col((p1.clone(), db::package::Column::Arch))
                        .is_in([arch, crate::ANY_ARCH]),
                )
                .add(
                    Expr::col((p1.clone(), db::package::Column::State))
                        .ne(db::PackageState::PendingDeletion),
                ),
        )
        .to_owned();
    let builder = conn.get_database_backend();
    let sql = builder.build(&query);

    db::Package::find().from_raw_sql(sql)
}

fn stale_pkgs_query(include_repo: bool) -> SelectStatement {
    let (p1, p2) = (Alias::new("p1"), Alias::new("p2"));
    let mut query = Query::select()
        .from_as(db::package::Entity, p1.clone())
        .to_owned();

    if include_repo {
        query.columns([
            (p1.clone(), db::package::Column::RepoId),
            (p1.clone(), db::package::Column::Id),
        ]);
    } else {
        query.column((p1.clone(), db::package::Column::Id));
    }

    // We left join on the max pkgs query because a repository that has all its packages set to
    // "pending deletion" doesn't show up in the query. These are also included with a where clause
    // on the joined rows.
    query
        .join_subquery(
            JoinType::LeftJoin,
            max_pkg_ids_query(true),
            p2.clone(),
            Condition::all()
                .add(
                    Expr::col((p1.clone(), db::package::Column::RepoId))
                        .eq(Expr::col((p2.clone(), db::package::Column::RepoId))),
                )
                .add(
                    Expr::col((p1.clone(), db::package::Column::Arch))
                        .eq(Expr::col((p2.clone(), db::package::Column::Arch))),
                )
                .add(
                    Expr::col((p1.clone(), db::package::Column::Name))
                        .eq(Expr::col((p2.clone(), db::package::Column::Name))),
                ),
        )
        .cond_where(
            Condition::any()
                .add(
                    Expr::col((p1.clone(), db::package::Column::Id))
                        .lt(Expr::col((p2.clone(), Alias::new("max_id")))),
                )
                .add(
                    Expr::col((p1.clone(), db::package::Column::State))
                        .eq(db::PackageState::PendingDeletion),
                ),
        );

    query
}

pub fn stale_pkgs(conn: &DbConn) -> SelectorRaw<SelectModel<PkgToRemove>> {
    let query = stale_pkgs_query(true);
    let builder = conn.get_database_backend();
    let sql = builder.build(&query);

    PkgToRemove::find_by_statement(sql)
}

pub async fn delete_stale_pkgs(conn: &DbConn, max_id: i32) -> crate::Result<()> {
    Ok(db::Package::delete_many()
        .filter(db::package::Column::Id.lte(max_id))
        .filter(db::package::Column::Id.in_subquery(stale_pkgs_query(false)))
        .exec(conn)
        .await
        .map(|_| ())?)
}
