use super::{archive, package, Command, SharedState};
use crate::db;

use std::{
    path::PathBuf,
    sync::{atomic::Ordering, Arc},
};

use futures::StreamExt;
use sea_orm::{ColumnTrait, EntityTrait, QueryFilter, QuerySelect};
use sea_query::Expr;
use tokio::{runtime, sync::mpsc};
use uuid::Uuid;

/// The actor is responsible for mutating the repositories. They receive their commands their
/// messages and process these commands in both a synchronous and asynchronous way.
pub struct Actor {
    rt: runtime::Handle,
    state: Arc<SharedState>,
}

impl Actor {
    pub fn new(rt: runtime::Handle, state: Arc<SharedState>) -> Self {
        Self {
            rt,
            state: Arc::clone(&state),
        }
    }

    pub fn random_file_paths<const C: usize>(&self) -> [PathBuf; C] {
        std::array::from_fn(|_| {
            let uuid: uuid::fmt::Simple = Uuid::new_v4().into();
            self.state.repos_dir.join(uuid.to_string())
        })
    }

    /// Run the main actor loop
    pub fn run(self) {
        while let Some(msg) = {
            let mut rx = self.state.rx.lock().unwrap();
            rx.blocking_recv()
        } {
            match msg {
                Command::ParsePkg(repo, path) => {
                    let _ = self.parse_pkg(repo, path);

                    if self
                        .state
                        .repos
                        .blocking_read()
                        .get(&repo)
                        .map(|n| n.0.load(Ordering::SeqCst))
                        == Some(0)
                    {
                        let _ = self.sync_repo(repo);
                        let _ = self.clean();
                    }
                }
                Command::SyncRepo(repo) => {
                    let _ = self.sync_repo(repo);
                }
                Command::Clean => {
                    let _ = self.clean();
                }
            }
        }
    }

    /// Parse a queued package for the given repository.
    fn parse_pkg(&self, repo: i32, path: PathBuf) -> crate::Result<()> {
        let pkg = package::Package::open(&path)?;
        let pkg = self
            .rt
            .block_on(db::query::package::insert(&self.state.conn, repo, pkg))?;
        let dest_path = self
            .state
            .repos_dir
            .join(repo.to_string())
            .join(pkg.id.to_string());
        std::fs::rename(path, dest_path)?;

        tracing::info!(
            "Added '{}-{}-{}' to repository {}",
            pkg.name,
            pkg.version,
            pkg.arch,
            repo,
        );

        self.state.repos.blocking_read().get(&repo).inspect(|n| {
            n.0.fetch_sub(1, Ordering::SeqCst);
        });

        Ok(())
    }

    fn sync_repo(&self, repo: i32) -> crate::Result<()> {
        let repos = self.state.repos.blocking_read();

        if let Some(_guard) = repos.get(&repo).map(|n| n.1.lock()) {
            let archs: Vec<String> = self.rt.block_on(
                db::Package::find()
                    .filter(db::package::Column::RepoId.eq(repo))
                    .select_only()
                    .column(db::package::Column::Arch)
                    .distinct()
                    .into_tuple()
                    .all(&self.state.conn),
            )?;

            for arch in archs {
                self.generate_archives(repo, &arch)?;
            }
        }

        Ok(())
    }

    fn generate_archives(&self, repo: i32, arch: &str) -> crate::Result<()> {
        let [tmp_ar_db_path, tmp_ar_files_path] = self.random_file_paths();

        let mut ars = archive::RepoArchivesWriter::new(
            &tmp_ar_db_path,
            &tmp_ar_files_path,
            self.random_file_paths(),
            &self.rt,
            &self.state.conn,
        )?;

        let (tx, mut rx) = mpsc::channel(1);

        let conn = self.state.conn.clone();
        let query = db::query::package::pkgs_to_sync(&self.state.conn, repo, arch);

        // sea_orm needs its connections to be dropped inside an async context, so we spawn a task
        // that streams the responses to the synchronous context via message passing
        self.rt.spawn(async move {
            match query.stream(&conn).await {
                Ok(mut stream) => {
                    while let Some(res) = stream.next().await {
                        let is_err = res.is_err();
                        let _ = tx.send(res).await;

                        if is_err {
                            return;
                        }
                    }
                }
                Err(err) => {
                    let _ = tx.send(Err(err)).await;
                }
            }
        });

        let mut committed_ids: Vec<i32> = Vec::new();

        while let Some(pkg) = rx.blocking_recv().transpose()? {
            committed_ids.push(pkg.id);
            ars.append_pkg(&pkg)?;
        }

        ars.close()?;

        // Move newly generated package archives to their correct place
        let repo_dir = self.state.repos_dir.join(repo.to_string());
        std::fs::rename(tmp_ar_db_path, repo_dir.join(format!("{}.db.tar.gz", arch)))?;
        std::fs::rename(
            tmp_ar_files_path,
            repo_dir.join(format!("{}.files.tar.gz", arch)),
        )?;

        // Update the state for the newly committed packages
        self.rt.block_on(
            db::Package::update_many()
                .col_expr(
                    db::package::Column::State,
                    Expr::value(db::PackageState::Committed),
                )
                .filter(db::package::Column::Id.is_in(committed_ids))
                .exec(&self.state.conn),
        )?;

        tracing::info!("Package archives generated for repo {} ('{}')", repo, arch);

        Ok(())
    }

    fn clean(&self) -> crate::Result<()> {
        let (tx, mut rx) = mpsc::channel(1);
        let conn = self.state.conn.clone();
        let query = db::query::package::stale_pkgs(&self.state.conn);

        // sea_orm needs its connections to be dropped inside an async context, so we spawn a task
        // that streams the responses to the synchronous context via message passing
        self.rt.spawn(async move {
            match query.stream(&conn).await {
                Ok(mut stream) => {
                    while let Some(res) = stream.next().await {
                        let is_err = res.is_err();
                        let _ = tx.send(res).await;

                        if is_err {
                            return;
                        }
                    }
                }
                Err(err) => {
                    let _ = tx.send(Err(err)).await;
                }
            }
        });

        // Ids are monotonically increasing, so the max id suffices to know which packages to
        // remove later
        let mut max_id = -1;
        let mut removed_pkgs = 0;

        while let Some(pkg) = rx.blocking_recv().transpose()? {
            // Failing to remove the package file isn't the biggest problem
            let _ = std::fs::remove_file(
                self.state
                    .repos_dir
                    .join(pkg.repo_id.to_string())
                    .join(pkg.id.to_string()),
            );

            if pkg.id > max_id {
                max_id = pkg.id;
            }

            removed_pkgs += 1;
        }

        if removed_pkgs > 0 {
            self.rt.block_on(db::query::package::delete_stale_pkgs(
                &self.state.conn,
                max_id,
            ))?;
        }

        tracing::info!("Cleaned up {removed_pkgs} old package(s)");

        Ok(())
    }
}
