use crate::db;
use std::{
    io::Write,
    path::{Path, PathBuf},
};

use futures::StreamExt;
use libarchive::{
    write::{Builder, FileWriter, WriteEntry},
    Entry, WriteFilter, WriteFormat,
};
use sea_orm::{ColumnTrait, DbConn, ModelTrait, QueryFilter, QuerySelect};
use tokio::{runtime, sync::mpsc};

pub struct RepoArchivesWriter {
    ar_db: FileWriter,
    ar_files: FileWriter,
    rt: runtime::Handle,
    conn: DbConn,
    tmp_paths: [PathBuf; 2],
}

impl RepoArchivesWriter {
    pub fn new(
        ar_db_path: impl AsRef<Path>,
        ar_files_path: impl AsRef<Path>,
        tmp_paths: [impl AsRef<Path>; 2],
        rt: &runtime::Handle,
        conn: &sea_orm::DbConn,
    ) -> crate::Result<Self> {
        let ar_db = Self::open_ar(ar_db_path)?;
        let ar_files = Self::open_ar(ar_files_path)?;

        Ok(Self {
            ar_db,
            ar_files,
            rt: rt.clone(),
            conn: conn.clone(),
            tmp_paths: [
                tmp_paths[0].as_ref().to_path_buf(),
                tmp_paths[1].as_ref().to_path_buf(),
            ],
        })
    }

    fn open_ar(path: impl AsRef<Path>) -> crate::Result<FileWriter> {
        let mut builder = Builder::new();
        builder.add_filter(WriteFilter::Gzip)?;
        builder.set_format(WriteFormat::PaxRestricted)?;

        Ok(builder.open_file(path)?)
    }

    fn append_entry(
        ar: &mut FileWriter,
        src_path: impl AsRef<Path>,
        dest_path: impl AsRef<Path>,
    ) -> crate::Result<()> {
        let metadata = std::fs::metadata(&src_path)?;
        let file_size = metadata.len();

        let mut ar_entry = WriteEntry::new();
        ar_entry.set_filetype(libarchive::archive::FileType::RegularFile);

        ar_entry.set_pathname(dest_path);
        ar_entry.set_mode(0o100644);
        ar_entry.set_size(file_size.try_into().unwrap());

        Ok(ar.append_path(&mut ar_entry, src_path)?)
    }

    pub fn append_pkg(&mut self, pkg: &db::package::Model) -> crate::Result<()> {
        self.write_desc(&self.tmp_paths[0], pkg)?;
        self.write_files(&self.tmp_paths[1], pkg)?;

        let full_name = format!("{}-{}", pkg.name, pkg.version);
        let dest_desc_path = format!("{}/desc", full_name);
        let dest_files_path = format!("{}/files", full_name);

        Self::append_entry(&mut self.ar_db, &self.tmp_paths[0], &dest_desc_path)?;
        Self::append_entry(&mut self.ar_files, &self.tmp_paths[0], &dest_desc_path)?;
        Self::append_entry(&mut self.ar_files, &self.tmp_paths[1], &dest_files_path)?;

        Ok(())
    }

    /// Generate a "files" archive entry for the package in the given path
    fn write_files(&self, path: impl AsRef<Path>, pkg: &db::package::Model) -> crate::Result<()> {
        let mut f = std::io::BufWriter::new(std::fs::File::create(path)?);

        writeln!(f, "%FILES%")?;

        let (tx, mut rx) = mpsc::channel(1);

        let conn = self.conn.clone();
        let query = pkg.find_related(db::PackageFile);

        self.rt.spawn(async move {
            match query.stream(&conn).await {
                Ok(mut stream) => {
                    while let Some(res) = stream.next().await {
                        let is_err = res.is_err();
                        let _ = tx.send(res).await;

                        if is_err {
                            return;
                        }
                    }
                }
                Err(err) => {
                    let _ = tx.send(Err(err)).await;
                }
            }
        });

        while let Some(file) = rx.blocking_recv().transpose()? {
            writeln!(f, "{}", file.path)?;
        }

        f.flush()?;
        Ok(())
    }

    fn write_desc(&self, path: impl AsRef<Path>, pkg: &db::package::Model) -> crate::Result<()> {
        let mut f = std::io::BufWriter::new(std::fs::File::create(path)?);

        let filename = format!(
            "{}-{}-{}.pkg.tar.{}",
            pkg.name, pkg.version, pkg.arch, pkg.compression
        );
        writeln!(f, "%FILENAME%\n{}", filename)?;

        let mut write_attr = |k: &str, v: &str| {
            if !v.is_empty() {
                writeln!(f, "\n%{}%\n{}", k, v)
            } else {
                Ok(())
            }
        };

        write_attr("NAME", &pkg.name)?;
        write_attr("BASE", &pkg.base)?;
        write_attr("VERSION", &pkg.version)?;

        if let Some(ref desc) = pkg.description {
            write_attr("DESC", desc)?;
        }

        let groups: Vec<String> = self.rt.block_on(
            pkg.find_related(db::PackageGroup)
                .select_only()
                .column(db::package_group::Column::Name)
                .into_tuple()
                .all(&self.conn),
        )?;

        write_attr("GROUPS", &groups.join("\n"))?;

        write_attr("CSIZE", &pkg.c_size.to_string())?;
        write_attr("ISIZE", &pkg.size.to_string())?;
        write_attr("SHA256SUM", &pkg.sha256_sum)?;

        if let Some(ref url) = pkg.url {
            write_attr("URL", url)?;
        }

        let licenses: Vec<String> = self.rt.block_on(
            pkg.find_related(db::PackageLicense)
                .select_only()
                .column(db::package_license::Column::Name)
                .into_tuple()
                .all(&self.conn),
        )?;
        write_attr("LICENSE", &licenses.join("\n"))?;

        write_attr("ARCH", &pkg.arch)?;

        // TODO build date
        write_attr(
            "BUILDDATE",
            &pkg.build_date.and_utc().timestamp().to_string(),
        )?;

        if let Some(ref packager) = pkg.packager {
            write_attr("PACKAGER", packager)?;
        }

        let related = [
            ("REPLACES", db::PackageRelatedEnum::Replaces),
            ("CONFLICTS", db::PackageRelatedEnum::Conflicts),
            ("PROVIDES", db::PackageRelatedEnum::Provides),
            ("DEPENDS", db::PackageRelatedEnum::Depend),
            ("OPTDEPENDS", db::PackageRelatedEnum::Optdepend),
            ("MAKEDEPENDS", db::PackageRelatedEnum::Makedepend),
            ("CHECKDEPENDS", db::PackageRelatedEnum::Checkdepend),
        ];

        for (key, attr) in related.into_iter() {
            let items: Vec<String> = self.rt.block_on(
                pkg.find_related(db::PackageRelated)
                    .filter(db::package_related::Column::Type.eq(attr))
                    .select_only()
                    .column(db::package_related::Column::Name)
                    .into_tuple()
                    .all(&self.conn),
            )?;

            write_attr(key, &items.join("\n"))?;
        }

        f.flush()?;
        Ok(())
    }

    pub fn close(&mut self) -> crate::Result<()> {
        self.ar_db.close()?;
        self.ar_files.close()?;

        let _ = std::fs::remove_file(&self.tmp_paths[0])?;
        let _ = std::fs::remove_file(&self.tmp_paths[1])?;

        Ok(())
    }
}
