mod actor;
mod archive;
mod handle;
pub mod package;

pub use actor::Actor;
pub use handle::Handle;

use crate::db;

use std::{
    collections::HashMap,
    path::{Path, PathBuf},
    sync::{atomic::AtomicU32, Arc, Mutex},
};

use sea_orm::{DbConn, EntityTrait, QuerySelect};
use tokio::{
    runtime,
    sync::{
        mpsc::{unbounded_channel, UnboundedReceiver, UnboundedSender},
        RwLock,
    },
};

pub enum Command {
    ParsePkg(i32, PathBuf),
    SyncRepo(i32),
    Clean,
}

pub struct SharedState {
    pub repos_dir: PathBuf,
    pub conn: DbConn,
    pub rx: Mutex<UnboundedReceiver<Command>>,
    pub tx: UnboundedSender<Command>,
    pub repos: RwLock<HashMap<i32, (AtomicU32, Arc<Mutex<()>>)>>,
}

impl SharedState {
    pub fn new(
        repos_dir: impl AsRef<Path>,
        conn: DbConn,
        repos: HashMap<i32, (AtomicU32, Arc<Mutex<()>>)>,
    ) -> Self {
        let (tx, rx) = unbounded_channel();

        Self {
            repos_dir: repos_dir.as_ref().to_path_buf(),
            conn,
            rx: Mutex::new(rx),
            tx,
            repos: RwLock::new(repos),
        }
    }
}

pub fn start(
    repos_dir: impl AsRef<Path>,
    conn: DbConn,
    rt: runtime::Handle,
    actors: u32,
) -> crate::Result<Handle> {
    std::fs::create_dir_all(repos_dir.as_ref())?;

    let mut repos = HashMap::new();
    let repo_ids: Vec<i32> = rt.block_on(
        db::Repo::find()
            .select_only()
            .column(db::repo::Column::Id)
            .into_tuple()
            .all(&conn),
    )?;

    for id in repo_ids {
        repos.insert(id, Default::default());
    }

    let state = Arc::new(SharedState::new(repos_dir, conn, repos));

    for _ in 0..actors {
        let actor = Actor::new(rt.clone(), Arc::clone(&state));

        std::thread::spawn(|| actor.run());
    }

    Ok(Handle::new(&state))
}
